"""Top-level package for PyGeoOGC."""
import os
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

from pygeoogc.cache_keys import create_request_key
from pygeoogc.exceptions import (
    InputTypeError,
    InputValueError,
    MissingInputError,
    ServiceError,
    ServiceUnavailableError,
    ZeroMatchedError,
)
from pygeoogc.print_versions import show_versions
from pygeoogc.pygeoogc import WFS, WMS, ArcGISRESTful, ServiceURL
from pygeoogc.utils import RetrySession, match_crs, streaming_download, traverse_json, validate_crs

cert_path = os.getenv("HYRIVER_SSL_CERT")
if cert_path is not None:
    from pyproj.network import set_ca_bundle_path

    if not Path(cert_path).exists():
        raise FileNotFoundError(cert_path)
    set_ca_bundle_path(cert_path)

try:
    __version__ = version("pygeoogc")
except PackageNotFoundError:
    __version__ = "999"

__all__ = [
    "ArcGISRESTful",
    "WFS",
    "WMS",
    "ServiceURL",
    "RetrySession",
    "InputTypeError",
    "InputValueError",
    "MissingInputError",
    "ServiceError",
    "ServiceUnavailableError",
    "ZeroMatchedError",
    "create_request_key",
    "traverse_json",
    "streaming_download",
    "match_crs",
    "validate_crs",
    "show_versions",
]



"""Base classes and function for REST, WMS, and WMF services."""

from __future__ import annotations

import itertools
import uuid
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator, Literal, Sequence, Union, cast, overload

import cytoolz.curried as tlz
import pyproj
from shapely import ops

import async_retriever as ar
from pygeoogc import utils
from pygeoogc.core import ArcGISRESTfulBase, WFSBase, WMSBase
from pygeoogc.exceptions import InputTypeError, InputValueError, ServiceError, ZeroMatchedError

if TYPE_CHECKING:
    from shapely import LineString, MultiPoint, MultiPolygon, Point, Polygon

    RESPONSE = Union[
        "list[str]", "list[bytes]", "list[dict[str, Any]]", "list[list[dict[str, Any]]]"
    ]
    CRSTYPE = Union[int, str, pyproj.CRS]


class ArcGISRESTful:
    """Access to an ArcGIS REST service.

    Notes
    -----
    By default, all retrieval methods retry to get the missing feature IDs,
    if there are any. You can disable this behavior by setting ``disable_retry``
    to ``True``. If there are any missing feature IDs after the retry,
    they are saved to a text file, ipath of which can be accessed by
    ``self.client.failed_path``.

    Parameters
    ----------
    base_url : str, optional
        The ArcGIS RESTful service url. The URL must either include a layer number
        after the last ``/`` in the url or the target layer must be passed as an argument.
    layer : int, optional
        Target layer number, defaults to None. If None layer number must be included as after
        the last ``/`` in ``base_url``.
    outformat : str, optional
        One of the output formats offered by the selected layer. If not correct
        a list of available formats is shown, defaults to ``geojson``.
    outfields : str or list
        The output fields to be requested. Setting ``*`` as outfields requests
        all the available fields which is the default behaviour.
    crs : str, int, or pyproj.CRS, optional
        The spatial reference of the output data, defaults to ``epsg:4326``.
    max_workers : int, optional
        Number of simultaneous download, default to 1, i.e., no threading. Note
        that some services might face issues when several requests are sent
        simultaneously and will return the requests partially. It's recommended
        to avoid using too many workers unless you are certain the web service
        can handle it.
    verbose : bool, optional
        If True, prints information about the requests and responses,
        defaults to False.
    disable_retry : bool, optional
        If ``True`` in case there are any failed queries, no retrying attempts
        is done and object IDs of the failed requests is saved to a text file
        which its ipath can be accessed via ``self.client.failed_path``.
    """

    def __init__(
        self,
        base_url: str,
        layer: int | None = None,
        outformat: str = "geojson",
        outfields: list[str] | str = "*",
        crs: CRSTYPE = 4326,
        max_workers: int = 1,
        verbose: bool = False,
        disable_retry: bool = False,
    ) -> None:
        self.client = ArcGISRESTfulBase(
            base_url=base_url,
            layer=layer,
            outformat=outformat,
            outfields=outfields,
            crs=crs,
            max_workers=max_workers,
            verbose=verbose,
            disable_retry=disable_retry,
        )

    def oids_bygeom(
        self,
        geom: (
            LineString
            | Polygon
            | Point
            | MultiPoint
            | tuple[float, float]
            | list[tuple[float, float]]
            | tuple[float, float, float, float]
        ),
        geo_crs: CRSTYPE = 4326,
        spatial_relation: str = "esriSpatialRelIntersects",
        sql_clause: str | None = None,
        distance: int | None = None,
    ) -> Iterator[tuple[str, ...]]:
        """Get feature IDs within a geometry that can be combined with a SQL where clause.

        Parameters
        ----------
        geom : LineString, Polygon, Point, MultiPoint, tuple, or list of tuples
            A geometry (LineString, Polygon, Point, MultiPoint), tuple of length two
            (``(x, y)``), a list of tuples of length 2 (``[(x, y), ...]``), or bounding box
            (tuple of length 4 (``(xmin, ymin, xmax, ymax)``)).
        geo_crs : str, int, or pyproj.CRS, optional
            The spatial reference of the input geometry, defaults to ``epsg:4326``.
        spatial_relation : str, optional
            The spatial relationship to be applied on the input geometry
            while performing the query. If not correct a list of available options is shown.
            It defaults to ``esriSpatialRelIntersects``. Valid predicates are:

            * ``esriSpatialRelIntersects``
            * ``esriSpatialRelContains``
            * ``esriSpatialRelCrosses``
            * ``esriSpatialRelEnvelopeIntersects``
            * ``esriSpatialRelIndexIntersects``
            * ``esriSpatialRelOverlaps``
            * ``esriSpatialRelTouches``
            * ``esriSpatialRelWithin``
            * ``esriSpatialRelRelation``

        sql_clause : str, optional
            Valid SQL 92 WHERE clause, default to None.
        distance : int, optional
            Buffer distance in meters for the input geometries, default to None.

        Returns
        -------
        list of tuples
            A list of feature IDs partitioned by ``self.max_nrecords``.
        """
        valid_spatialrels = [
            "esriSpatialRelIntersects",
            "esriSpatialRelContains",
            "esriSpatialRelCrosses",
            "esriSpatialRelEnvelopeIntersects",
            "esriSpatialRelIndexIntersects",
            "esriSpatialRelOverlaps",
            "esriSpatialRelTouches",
            "esriSpatialRelWithin",
            "esriSpatialRelRelation",
        ]
        if spatial_relation not in valid_spatialrels:
            raise InputValueError("spatial_relation", valid_spatialrels)

        geom_query = utils.esri_query(geom, geo_crs, self.client.crs)
        payload = {
            **geom_query,
            "spatialRel": spatial_relation,
            "returnGeometry": "false",
            "returnIdsOnly": "true",
            "f": self.client.outformat,
        }
        if distance:
            payload.update({"distance": str(int(distance)), "units": "esriSRUnit_Meter"})

        if sql_clause:
            payload.update({"where": sql_clause})

        self.client.request_id = uuid.uuid4().hex

        resp = self.client.get_response(self.client.query_url, [payload], method="POST")[0]
        try:
            return self.partition_oids(resp["objectIds"])
        except (KeyError, TypeError) as ex:
            msg = resp["error"]["message"] if "error" in resp else "No matched records"
            raise ZeroMatchedError(msg) from ex

    def oids_byfield(
        self, field: str, ids: str | int | list[str] | list[int]
    ) -> Iterator[tuple[str, ...]]:
        """Get Object IDs based on a list of field IDs.

        Parameters
        ----------
        field : str
            Name of the target field that IDs belong to.
        ids : str or list
            A list of target ID(s).

        Returns
        -------
        list of tuples
            A list of feature IDs partitioned by ``self.max_nrecords``.
        """
        if field not in self.client.valid_fields:
            raise InputValueError("field", self.client.valid_fields)

        ids_ls = {ids} if isinstance(ids, (str, int)) else set(ids)

        if "string" in self.client.field_types.get(field, ""):
            fids = ", ".join(f"'{i}'" for i in ids_ls)
        else:
            fids = ", ".join(str(i) for i in ids_ls)

        return self.oids_bysql(f"{field} IN ({fids})")

    def oids_bysql(self, sql_clause: str) -> Iterator[tuple[str, ...]]:
        """Get feature IDs using a valid SQL 92 WHERE clause.

        Notes
        -----
        Not all web services support this type of query. For more details look
        `here <https://developers.arcgis.com/rest/services-reference/query-feature-service-.htm#ESRI_SECTION2_07DD2C5127674F6A814CE6C07D39AD46>`__.

        Parameters
        ----------
        sql_clause : str
            A valid SQL 92 WHERE clause.

        Returns
        -------
        list of tuples
            A list of feature IDs partitioned by ``self.max_nrecords``.
        """
        if not isinstance(sql_clause, str):
            raise InputTypeError("sql_clause", "str")

        payload = {
            "where": sql_clause,
            "returnGeometry": "false",
            "returnIdsOnly": "true",
            "f": self.client.outformat,
        }
        self.client.request_id = uuid.uuid4().hex

        resp = self.client.get_response(self.client.query_url, [payload], method="POST")[0]
        try:
            return self.partition_oids(resp["objectIds"])
        except (KeyError, TypeError) as ex:
            msg = resp["error"]["message"] if "error" in resp else "No matched records"
            raise ZeroMatchedError(msg) from ex

    def partition_oids(self, oids: list[int] | int) -> Iterator[tuple[str, ...]]:
        """Partition feature IDs based on ``self.max_nrecords``.

        Parameters
        ----------
        oids : list of int or int
            A list of feature ID(s).

        Returns
        -------
        list of tuples
            A list of feature IDs partitioned by ``self.max_nrecords``.
        """
        return self.client.partition_oids(oids)

    def get_features(
        self,
        featureids: Iterator[tuple[str, ...]],
        return_m: bool = False,
        return_geom: bool = True,
    ) -> list[dict[str, Any]]:
        """Get features based on the feature IDs.

        Parameters
        ----------
        featureids : list
            List of feature IDs.
        return_m : bool, optional
            Whether to activate the Return M (measure) in the request,
            defaults to ``False``.
        return_geom : bool, optional
            Whether to return the geometry of the feature, defaults to ``True``.

        Returns
        -------
        dict
            (Geo)json response from the web service.
        """
        return self.client.get_features(featureids, return_m, return_geom)

    def __repr__(self) -> str:
        """Print the service configuration."""
        return self.client.__repr__()


class WMS:
    """Get data from a WMS service within a geometry or bounding box.

    Parameters
    ----------
    url : str
        The base url for the WMS service e.g., https://www.mrlc.gov/geoserver/mrlc_download/wms
    layers : str or list
        A layer or a list of layers from the service to be downloaded. You can pass an empty
        string to get a list of available layers.
    outformat : str
        The data format to request for data from the service. You can pass an empty
        string to get a list of available output formats.
    crs : str, int, or pyproj.CRS, optional
        The spatial reference system to be used for requesting the data, defaults to
        ``epsg:4326``.
    version : str, optional
        The WMS service version which should be either 1.1.1 or 1.3.0, defaults to 1.3.0.
    validation : bool, optional
        Validate the input arguments from the WMS service, defaults to True. Set this
        to False if you are sure all the WMS settings such as layer and crs are correct
        to avoid sending extra requests.
    ssl : bool, optional
        Whether to use SSL for the connection, defaults to ``True``.
    """

    def __init__(
        self,
        url: str,
        layers: str | list[str],
        outformat: str,
        version: str = "1.3.0",
        crs: CRSTYPE = 4326,
        validation: bool = True,
        ssl: bool = True,
    ) -> None:
        self.client = WMSBase(
            url=url,
            layers=layers,
            outformat=outformat,
            version=version,
            crs=crs,
            validation=validation,
        )
        self.url = self.client.url
        self.outformat = self.client.outformat
        self.version = self.client.version
        self.crs = self.client.crs
        self.crs_str = self.client.crs_str
        self.is_geographic = pyproj.CRS(self.crs).is_geographic
        self.ssl = ssl
        self.layers = self.client.layers

    def get_validlayers(self) -> dict[str, str]:
        """Get the layers supported by the WMS service."""
        return self.client.get_validlayers()

    @overload
    def getmap_bybox(
        self,
        bbox: tuple[float, float, float, float],
        resolution: float,
        box_crs: CRSTYPE = ...,
        always_xy: bool = ...,
        max_px: int = ...,
        kwargs: dict[str, Any] | None = ...,
        tiff_dir: Literal[None] = None,
    ) -> dict[str, bytes]:
        ...

    @overload
    def getmap_bybox(
        self,
        bbox: tuple[float, float, float, float],
        resolution: float,
        box_crs: CRSTYPE = ...,
        always_xy: bool = ...,
        max_px: int = ...,
        kwargs: dict[str, Any] | None = ...,
        tiff_dir: str | Path = ...,
    ) -> list[Path]:
        ...

    def getmap_bybox(
        self,
        bbox: tuple[float, float, float, float],
        resolution: float,
        box_crs: CRSTYPE = 4326,
        always_xy: bool = False,
        max_px: int = 8000000,
        kwargs: dict[str, Any] | None = None,
        tiff_dir: str | Path | None = None,
    ) -> dict[str, bytes] | list[Path]:
        """Get data from a WMS service within a geometry or bounding box.

        Parameters
        ----------
        bbox : tuple
            A bounding box for getting the data.
        resolution : float
            The output resolution in meters. The width and height of output are computed in pixel
            based on the geometry bounds and the given resolution.
        box_crs : str, int, or pyproj.CRS, optional
            The spatial reference system of the input bbox, defaults to
            ``epsg:4326``.
        always_xy : bool, optional
            Whether to always use xy axis order, defaults to False. Some services change the axis
            order from xy to yx, following the latest WFS version specifications but some don't.
            If the returned value does not have any geometry, it indicates that most probably the
            axis order does not match. You can set this to True in that case.
        max_px : int, optional
            The maximum allowable number of pixels (width x height) for a WMS requests,
            defaults to 8 million based on some trial-and-error.
        kwargs: dict, optional
            Optional additional keywords passed as payload, defaults to None.
            For example, ``{"styles": "default"}``.
        tiff_dir : str or pathlib.Path, optional
            If given, the retrieved data will be stored on disk instead of
            returning it, defaults to ``None``, i.e., saving to memory
            and returning the data.

        Returns
        -------
        dict of bytes or list of pathlib.Path
            If ``to_disk=False``, a dict where the keys are the layer name and
            values are the returned response from the WMS service as bytes.
            If ``to_disk=True``, a list of pathlib.Path objects to the saved files.
        """
        utils.check_bbox(bbox)
        _bbox = utils.match_crs(bbox, box_crs, self.crs_str)
        bounds = utils.bbox_decompose(_bbox, resolution, self.crs_str, max_px)

        payload = {
            "version": self.version,
            "format": self.outformat,
            "request": "GetMap",
        }

        if kwargs is not None and not isinstance(kwargs, dict):
            raise InputTypeError("kwargs", "dict or None")

        if isinstance(kwargs, dict):
            payload.update(kwargs)

        if self.version == "1.1.1":
            payload["srs"] = self.crs_str
        else:
            payload["crs"] = self.crs_str

        def _get_payloads(
            args: tuple[str, tuple[tuple[float, float, float, float], str, int, int]]
        ) -> tuple[str, dict[str, str]]:
            lyr, bnds = args
            _bbox, counter, _width, _height = bnds

            if self.version != "1.1.1" and self.is_geographic and not always_xy:
                _bbox = (_bbox[1], _bbox[0], _bbox[3], _bbox[2])
            _payload = payload.copy()
            _payload["bbox"] = ",".join(str(round(c, 6)) for c in _bbox)
            _payload["width"] = str(_width)
            _payload["height"] = str(_height)
            _payload["layers"] = lyr
            return f"{lyr}_dd_{counter}", _payload

        _lyr_payloads = (_get_payloads(i) for i in itertools.product(self.layers, bounds))
        layers, payloads = zip(*_lyr_payloads)
        layers = cast("tuple[str]", layers)
        payloads = cast("tuple[dict[str, str]]", payloads)
        if tiff_dir is not None:
            tiff_dir = Path(tiff_dir)
            tiff_dir.mkdir(parents=True, exist_ok=True)
            return utils.streaming_download(
                [self.url] * len(payloads),
                [{"params": p} for p in payloads],
                root_dir=tiff_dir,
                file_extention="tiff",
                n_jobs=4,
                ssl=self.ssl,
            )
        rbinary = ar.retrieve_binary(
            [self.url] * len(payloads),
            [{"params": p} for p in payloads],
            max_workers=4,
            ssl=self.ssl,
        )
        return dict(zip(layers, rbinary))

    def __repr__(self) -> str:
        """Print the services properties."""
        return self.client.__repr__()


class WFS(WFSBase):
    """Data from any WFS service within a geometry or by featureid.

    Parameters
    ----------
    url : str
        The base url for the WFS service, for examples:
        https://hazards.fema.gov/nfhl/services/public/NFHL/MapServer/WFSServer
    layer : str
        The layer from the service to be downloaded, defaults to None which throws
        an error and includes all the available layers offered by the service.
    outformat : str
        The data format to request for data from the service, defaults to None which
         throws an error and includes all the available format offered by the service.
    version : str, optional
        The WFS service version which should be either 1.0.0, 1.1.0, or 2.0.0.
        Defaults to 2.0.0.
    crs : str, int, or pyproj.CRS, optional
        The spatial reference system to be used for requesting the data, defaults to
        ``epsg:4326``.
    read_method : str, optional
        Method for reading the retrieved data, defaults to ``json``. Valid options are
        ``json``, ``binary``, and ``text``.
    max_nrecords : int, optional
        The maximum number of records in a single request to be retrieved from the service,
        defaults to 1000. If the number of records requested is greater than this value,
        it will be split into multiple requests.
    validation : bool, optional
        Validate the input arguments from the WFS service, defaults to True. Set this
        to False if you are sure all the WFS settings such as layer and crs are correct
        to avoid sending extra requests.
    """

    def __init__(
        self,
        url: str,
        layer: str | None = None,
        outformat: str | None = None,
        version: str = "2.0.0",
        crs: CRSTYPE = 4326,
        read_method: str = "json",
        max_nrecords: int = 1000,
        validation: bool = True,
    ) -> None:
        super().__init__(
            url=url,
            layer=layer,
            outformat=outformat,
            version=version,
            crs=crs,
            read_method=read_method,
            max_nrecords=max_nrecords,
            validation=validation,
        )

    def getfeature_bybox(
        self,
        bbox: tuple[float, float, float, float],
        box_crs: CRSTYPE = 4326,
        always_xy: bool = False,
        sort_attr: str | None = None,
    ) -> RESPONSE:
        """Get data from a WFS service within a bounding box.

        Parameters
        ----------
        bbox : tuple
            A bounding box for getting the data: [west, south, east, north]
        box_crs : str, or pyproj.CRS, optional
            The spatial reference system of the input bbox, defaults to
            ``epsg:4326``.
        always_xy : bool, optional
            Whether to always use xy axis order, defaults to False. Some services change the axis
            order from xy to yx, following the latest WFS version specifications but some don't.
            If the returned value does not have any geometry, it indicates that most probably the
            axis order does not match. You can set this to True in that case.
        sort_attr : str, optional
            The column name in the database to sort request by, defaults
            to the first attribute in the schema that contains ``id`` in its name.

        Returns
        -------
        list of str or bytes or dict
            WFS query response within a bounding box.
        """
        utils.check_bbox(bbox)
        box_crs = pyproj.CRS(box_crs)

        if box_crs.is_geographic and not always_xy:
            bbox = (bbox[1], bbox[0], bbox[3], bbox[2])

        bbox_str = f'{",".join(str(round(c, 6)) for c in bbox)},{box_crs.to_string()}'
        payload = {
            "service": "wfs",
            "version": self.version,
            "outputFormat": "text/xml",
            "request": "GetFeature",
            "typeName": self.layer,
            "bbox": bbox_str,
            "srsName": self.crs_str,
            "resultType": "hits",
        }
        resp = ar.retrieve_text([self.url], [{"params": payload}])
        try:
            nfeatures = int(resp[0].split(self.nfeat_key)[-1].split(" ")[0].strip('"'))
        except (IndexError, ValueError) as ex:
            raise ServiceError(resp[0], self.url) from ex

        payloads = [
            {
                "service": "wfs",
                "version": self.version,
                "outputFormat": self.outformat,
                "request": "GetFeature",
                "typeName": self.layer,
                "bbox": bbox_str,
                "srsName": self.crs_str,
                **self.sort_params(sort_attr, nfeatures, i),
            }
            for i in range(0, nfeatures, self.max_nrecords)
        ]

        return self.retrieve([self.url] * len(payloads), [{"params": p} for p in payloads])

    def getfeature_bygeom(
        self,
        geometry: Polygon | MultiPolygon,
        geo_crs: CRSTYPE = 4326,
        always_xy: bool = False,
        predicate: str = "INTERSECTS",
        sort_attr: str | None = None,
    ) -> RESPONSE:
        """Get features based on a geometry.

        Parameters
        ----------
        geometry : shapely.geometry
            The input geometry
        geo_crs : str, or pyproj.CRS, optional
            The CRS of the input geometry, default to ``epsg:4326``.
        always_xy : bool, optional
            Whether to always use xy axis order, defaults to False. Some services change the axis
            order from xy to yx, following the latest WFS version specifications but some don't.
            If the returned value does not have any geometry, it indicates that most probably the
            axis order does not match. You can set this to True in that case.
        predicate : str, optional
            The geometric predicate to use for requesting the data, defaults to ``INTERSECTS``.
            Valid predicates are:

            * ``EQUALS``
            * ``DISJOINT``
            * ``INTERSECTS``
            * ``TOUCHES``
            * ``CROSSES``
            * ``WITHIN``
            * ``CONTAINS``
            * ``OVERLAPS``
            * ``RELATE``
            * ``BEYOND``

        sort_attr : str, optional
            The column name in the database to sort request by, defaults
            to the first attribute in the schema that contains ``id`` in its name.

        Returns
        -------
        str or bytes or dict
            WFS query response based on the given geometry.
        """
        geom = utils.match_crs(geometry, geo_crs, self.crs_str)
        geom_name = ""
        if "geometry_column" in self.schema[self.layer]:
            geom_name = self.schema[self.layer]["geometry_column"]
        elif "properties" in self.schema[self.layer]:
            geom_name = next(
                (lyr for lyr in self.schema[self.layer]["properties"] if "geom" in lyr), ""
            )
        elif "required" in self.schema[self.layer]:
            geom_name = next(
                (lyr for lyr in self.schema[self.layer]["required"] if "geom" in lyr), ""
            )

        if not geom_name:
            msg = "Cannot find the geometry column name in the schema."
            raise ValueError(msg)

        if pyproj.CRS(geo_crs).is_geographic and not always_xy:
            g_wkt = ops.transform(lambda x, y: (y, x), geom).wkt
        else:
            g_wkt = geom.wkt

        valid_predicates = [
            "EQUALS",
            "DISJOINT",
            "INTERSECTS",
            "TOUCHES",
            "CROSSES",
            "WITHIN",
            "CONTAINS",
            "OVERLAPS",
            "RELATE",
            "BEYOND",
        ]
        if predicate.upper() not in valid_predicates:
            raise InputValueError("predicate", valid_predicates)

        return self.getfeature_byfilter(
            f"{predicate.upper()}({geom_name}, {g_wkt})", method="POST", sort_attr=sort_attr
        )

    def getfeature_byid(
        self,
        featurename: str,
        featureids: Sequence[int | str] | int | str,
    ) -> RESPONSE:
        """Get features based on feature IDs.

        Parameters
        ----------
        featurename : str
            The name of the column for searching for feature IDs.
        featureids : int, str, or list of them
            The feature ID(s).

        Returns
        -------
        str or bytes or dict
            WMS query response.
        """
        valid_features = self.schema[self.layer]["properties"]
        if featurename not in valid_features:
            raise InputValueError("featurename", list(valid_features))

        feat_set = {featureids} if isinstance(featureids, (str, int)) else set(featureids)

        if not feat_set:
            raise InputTypeError("featureids", "int, str, or list")

        feat_itr = tlz.partition_all(self.max_nrecords, feat_set)
        if "str" in valid_features[featurename]:
            feat_vals = (", ".join(f"'{fid}'" for fid in fids) for fids in feat_itr)
        else:
            feat_vals = (", ".join(str(fid) for fid in fids) for fids in feat_itr)

        return list(
            tlz.concat(
                self.getfeature_byfilter(
                    f"{featurename} IN ({f})", method="POST", sort_attr=featurename
                )
                for f in feat_vals
            )
        )

    def getfeature_byfilter(
        self, cql_filter: str, method: str = "GET", sort_attr: str | None = None
    ) -> RESPONSE:
        """Get features based on a valid CQL filter.

        Notes
        -----
        The validity of the input CQL expression is user's responsibility since
        the function does not perform any checks and just sends a request using
        the input filter.

        Parameters
        ----------
        cql_filter : str
            A valid CQL filter expression.
        method : str
            The request method, could be GET or POST (for long filters).
        sort_attr : str, optional
            The column name in the database to sort request by, defaults
            to the first attribute in the schema that contains ``id`` in its name.

        Returns
        -------
        str or bytes or dict
            WFS query response
        """
        if not isinstance(cql_filter, str):
            raise InputTypeError("cql_filter", "str")

        valid_methods = ["GET", "POST"]
        if method not in valid_methods:
            raise InputValueError("method", valid_methods)

        payload = {
            "service": "wfs",
            "version": self.version,
            "outputFormat": "text/xml",
            "request": "GetFeature",
            "typeName": self.layer,
            "srsName": self.crs_str,
            "cql_filter": cql_filter,
            "resultType": "hits",
        }

        if method == "GET":
            resp = ar.retrieve_text([self.url], [{"params": payload}])
        else:
            headers = {"content-type": "application/x-www-form-urlencoded"}
            resp = ar.retrieve_text([self.url], [{"data": payload, "headers": headers}], "POST")
        try:
            nfeatures = int(resp[0].split(self.nfeat_key)[-1].split(" ")[0].strip('"'))
        except (IndexError, ValueError) as ex:
            raise ServiceError(resp[0], self.url) from ex

        payloads = [
            {
                "service": "wfs",
                "version": self.version,
                "outputFormat": self.outformat,
                "request": "GetFeature",
                "typeName": self.layer,
                "srsName": self.crs_str,
                "cql_filter": cql_filter,
                **self.sort_params(sort_attr, nfeatures, i),
            }
            for i in range(0, nfeatures, self.max_nrecords)
        ]
        if method == "GET":
            return self.retrieve([self.url] * len(payloads), [{"params": p} for p in payloads])

        headers = {"content-type": "application/x-www-form-urlencoded"}
        return self.retrieve(
            [self.url] * len(payloads),
            [{"data": p, "headers": headers} for p in payloads],
            "POST",
        )


@dataclass(frozen=True)
class HttpURLs:
    """URLs of the supported HTTP services."""

    ssebopeta: str = "https://edcintl.cr.usgs.gov/downloads/sciweb1/shared/uswem/web/conus/eta/modis_eta/daily/downloads"


@dataclass(frozen=True)
class RESTfulURLs:
    """URLs of the supported RESTful services."""

    daymet: str = "https://thredds.daac.ornl.gov/thredds/ncss/ornldaac"
    daymet_point: str = "https://daymet.ornl.gov/single-pixel/api/data"
    gridmet: str = "http://thredds.northwestknowledge.net:8080/thredds/ncss"
    fema_nfhl: str = "https://hazards.fema.gov/gis/nfhl/rest/services/public/NFHL/MapServer"
    fema_prelim_cslf: str = (
        "https://hazards.fema.gov/gis/nfhl/rest/services/CSLF/Prelim_CSLF/MapServer"
    )
    fema_draft_cslf: str = (
        "https://hazards.fema.gov/gis/nfhl/rest/services/CSLF/Draft_CSLF/MapServer"
    )
    fema_prelim_nfhl: str = (
        "https://hazards.fema.gov/gis/nfhl/rest/services/PrelimPending/Prelim_NFHL/MapServer"
    )
    fema_pending_nfhl: str = (
        "https://hazards.fema.gov/gis/nfhl/rest/services/PrelimPending/Pending_NFHL/MapServer"
    )
    fema_draft_nfhl: str = (
        "https://hazards.fema.gov/gis/nfhl/rest/services/AFHI/Draft_FIRM_DB/MapServer"
    )
    fws: str = "https://www.fws.gov/wetlandsmapservice/rest/services"
    nldi: str = "https://labs.waterdata.usgs.gov/api/nldi"
    nwis: str = "https://waterservices.usgs.gov/nwis"
    wbd: str = "https://hydro.nationalmap.gov/arcgis/rest/services/wbd/MapServer"
    nhd: str = "https://hydro.nationalmap.gov/arcgis/rest/services/nhd/MapServer"
    hp3d: str = "https://hydro.nationalmap.gov/arcgis/rest/services/3DHP_all/MapServer"
    nhdplushr: str = "https://hydro.nationalmap.gov/arcgis/rest/services/NHDPlus_HR/MapServer"
    nhdhr_edits: str = "https://edits.nationalmap.gov/arcgis/rest/services/HEM/NHDHigh/MapServer"
    nhdplushr_edits: str = (
        "https://edits.nationalmap.gov/arcgis/rest/services/NHDPlus_HR/NHDPlus_HR/MapServer"
    )
    nhdplus_epa: str = "https://watersgeo.epa.gov/arcgis/rest/services/NHDPlus/NHDPlus/MapServer"
    nhd_fabric: str = (
        "https://watersgeo.epa.gov/arcgis/rest/services/Support/CatchmentFabric/MapServer"
    )
    nid: str = "https://nid.sec.usace.army.mil/api"
    airmap: str = "https://api.airmap.com/elevation/v1/ele"
    nm_pqs: str = "https://epqs.nationalmap.gov/v1/json"
    pygeoapi: str = "https://labs.waterdata.usgs.gov/api/nldi/pygeoapi/processes"
    nm_3dep_index: str = (
        "https://index.nationalmap.gov/arcgis/rest/services/3DEPElevationIndex/MapServer"
    )
    geoconnex: str = "https://reference.geoconnex.us"
    stnflood: str = "https://stn.wim.usgs.gov/STNServices/"
    stnflood_dd: str = "https://stn.wim.usgs.gov/STNWeb/datadictionary/"
    ehydro: str = "https://services7.arcgis.com/n1YM8pTrFmm7L4hs/ArcGIS/rest/services/eHydro_Survey_Data/FeatureServer/0"
    ehydro_bins: str = "https://services7.arcgis.com/n1YM8pTrFmm7L4hs/ArcGIS/rest/services/Hydrographic_Survey_Bins/FeatureServer/2"


@dataclass(frozen=True)
class WFSURLs:
    """URLs of the supported WFS services."""

    fema: str = "https://hazards.fema.gov/gis/nfhl/services/public/NFHL/MapServer/WFSServer"
    waterdata: str = "https://labs.waterdata.usgs.gov/geoserver/wmadata/ows"


@dataclass(frozen=True)
class WMSURLs:
    """URLs of the supported WMS services."""

    fema: str = "https://hazards.fema.gov/gis/nfhl/rest/services/public/NFHLWMS/MapServer/WMSServer"
    fws: str = (
        "https://www.fws.gov/wetlandsmapservice/services/Wetlands_Raster/ImageServer/WMSServer"
    )
    mrlc: str = "https://www.mrlc.gov/geoserver/mrlc_download/wms"
    nm_3dep: str = (
        "https://elevation.nationalmap.gov/arcgis/services/3DEPElevation/ImageServer/WMSServer"
    )
    gebco: str = (
        "https://www.gebco.net/data_and_products/gebco_web_services/web_map_service/mapserv"
    )


@dataclass(frozen=True)
class ServiceURL:
    """URLs of the supported services."""

    http: HttpURLs = field(default_factory=HttpURLs)
    restful: RESTfulURLs = field(default_factory=RESTfulURLs)
    wfs: WFSURLs = field(default_factory=WFSURLs)
    wms: WMSURLs = field(default_factory=WMSURLs)



"""Base classes and function for REST, WMS, and WMF services."""
from __future__ import annotations

import contextlib
import math
import os
import uuid
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator, Literal, Sequence, Union, cast

import cytoolz.curried as tlz
import pyproj
from owslib.wfs import WebFeatureService
from owslib.wms import WebMapService
from yarl import URL

import async_retriever as ar
from pygeoogc import utils
from pygeoogc.exceptions import (
    InputTypeError,
    InputValueError,
    MissingInputError,
    ServiceError,
    ServiceUnavailableError,
    ZeroMatchedError,
)

if TYPE_CHECKING:
    CRSTYPE = Union[int, str, pyproj.CRS]

__all__ = ["ArcGISRESTfulBase", "WMSBase", "WFSBase"]


def _extract_layer(url: str, layer: int | None) -> tuple[str, int]:
    """Check if layer is included in url, if so separate and return them."""
    url_obj = URL(url[:-1] if url.endswith("/") else url)
    if layer is None:
        try:
            return url_obj.parent.human_repr(), int(url_obj.parts[-1])
        except ValueError as ex:
            msg = "Either layer must be passed as an argument or be included in ``base_url``"
            raise MissingInputError(msg) from ex
    return url_obj.human_repr(), layer


class ArcGISRESTfulBase:
    """Access to an ArcGIS REST service.

    Parameters
    ----------
    base_url : str, optional
        The ArcGIS RESTful service url. The URL must either include a layer number
        after the last ``/`` in the url or the target layer must be passed as an
        argument.
    layer : int, optional
        Target layer number, defaults to None. If None layer number must be
        included as after the last ``/`` in ``base_url``.
    outformat : str, optional
        One of the output formats offered by the selected layer. If not correct
        a list of available formats is shown, defaults to ``geojson``.
        It defaults to ``esriSpatialRelIntersects``.
    outfields : str or list
        The output fields to be requested. Setting ``*`` as outfields requests
        all the available fields which is the default setting.
    crs : str, int, or pyproj.CRS, optional
        The spatial reference of the output data, defaults to ``epsg:4326``
    max_workers : int, optional
        Max number of simultaneous requests, default to 2. Note
        that some services might face issues when several requests are sent
        simultaneously and will return the requests partially. It's recommended
        to avoid using too many workers unless you are certain the web service
        can handle it.
    verbose : bool, optional
        If True, prints information about the requests and responses,
        defaults to False.
    disable_retry : bool, optional
        If ``True`` in case there are any failed queries, no retrying attempts
        is done and object IDs of the failed requests is saved to a text file
        which its ipath can be accessed via ``self.failed_path``.
    """

    def __init__(
        self,
        base_url: str,
        layer: int | None = None,
        outformat: str = "geojson",
        outfields: list[str] | str = "*",
        crs: CRSTYPE = 4326,
        max_workers: int = 1,
        verbose: bool = False,
        disable_retry: bool = False,
    ) -> None:
        self.base_url, self.layer = _extract_layer(base_url, layer)
        self.outformat = outformat
        self.outfields = outfields if isinstance(outfields, (list, tuple)) else [outfields]
        self.crs = utils.validate_crs(crs)
        self.max_workers = max_workers
        self.verbose = verbose
        self.disable_retry = disable_retry

        self.out_sr = cast("int", pyproj.CRS(self.crs).to_epsg())
        self.n_features = 0
        self.url = f"{self.base_url}/{self.layer}"
        self.query_url = f"{self.url}/query"
        self.valid_layers: dict[str, str] = {}
        self.query_formats: list[str] = []
        self.extent: tuple[float, float, float, float] | None = None
        self.units: str | None = None
        self.max_nrecords = 1000
        self.valid_fields: list[str] = []
        self.field_types: dict[str, str] = {}
        self.feature_types: dict[int, str] | None = None
        self.return_m = False
        self.return_geom = True
        self.n_missing = 0
        self.total_n_features = 0
        self.failed_path = Path("cache", "failed_ids.txt")
        self.request_id: str | None = None

        self.initialize_service()

    def _set_service_properties(self) -> None:
        rjson = self.get_response(self.base_url, [{"f": "json"}])[0]
        try:
            self.valid_layers = {str(lyr["id"]): lyr["name"] for lyr in rjson["layers"]}
            self.query_formats = rjson["supportedQueryFormats"].replace(" ", "").lower().split(",")

            extent = rjson["extent"] if "extent" in rjson else rjson["fullExtent"]
            bounds = (extent["xmin"], extent["ymin"], extent["xmax"], extent["ymax"])
            if any(math.isnan(float(b)) for b in bounds):
                if rjson.get("initialExtent"):
                    extent = rjson["initialExtent"]
                    bounds = (extent["xmin"], extent["ymin"], extent["xmax"], extent["ymax"])
                else:
                    raise ServiceError(self.base_url)
        except (ValueError, KeyError) as ex:
            raise ServiceError(self.base_url) from ex

        crs_iter = iter(["latestWkid", "wkid", "wkt"])
        while True:
            try:
                crs = utils.validate_crs(extent["spatialReference"].get(next(crs_iter)))
                self.extent = utils.match_crs(bounds, crs, 4326)
                break
            except InputTypeError:
                continue
            except StopIteration as ex:
                raise ServiceError(self.base_url) from ex

        with contextlib.suppress(KeyError):
            self.units = rjson["units"].replace("esri", "").lower()
            self.max_nrecords = int(rjson["maxRecordCount"])

    def _set_layer_properties(self) -> None:
        """Set properties of the target layer."""
        rjson = self.get_response(self.url, [{"f": "json"}])[0]
        if "fields" not in rjson:
            return
        self.valid_fields = list(
            set(
                utils.traverse_json(rjson, ["fields", "name"])
                + utils.traverse_json(rjson, ["fields", "alias"])
                + ["*"]
            )
        )
        self.field_types = {
            f["name"]: f["type"].replace("esriFieldType", "").lower() for f in rjson["fields"]
        }
        if "alias" in rjson["fields"][0]:
            self.field_types.update(
                {
                    f["alias"]: f["type"].replace("esriFieldType", "").lower()
                    for f in rjson["fields"]
                }
            )

        with contextlib.suppress(KeyError):
            self.feature_types = dict(
                zip((tlz.pluck("id", rjson["types"])), tlz.pluck("name", rjson["types"]))
            )

    def initialize_service(self) -> None:
        """Initialize the RESTFul service."""
        self._set_service_properties()
        if str(self.layer) not in self.valid_layers:
            valids = [f'"{i}" for {n}' for i, n in self.valid_layers.items()]
            raise InputValueError("layer", valids)

        self._set_layer_properties()

        if self.outformat.lower() not in self.query_formats:
            raise InputValueError("outformat", self.query_formats)

        if self.valid_fields and any(f not in self.valid_fields for f in self.outfields):
            raise InputValueError("outfields", self.valid_fields)

    def partition_oids(self, oids: list[int] | int) -> Iterator[tuple[str, ...]]:
        """Partition feature IDs based on ``self.max_nrecords``."""
        oid_list = [oids] if isinstance(oids, int) else set(oids)
        if len(oid_list) == 0:
            raise ZeroMatchedError

        self.n_features = len(oid_list)
        return tlz.partition_all(self.max_nrecords, [str(i) for i in oid_list])

    def _cleanup_resp(
        self, resp: list[dict[str, Any]], payloads: list[dict[str, str]]
    ) -> list[dict[str, Any]]:
        """Remove failed responses."""
        fails = [i for i, r in enumerate(resp) if "error" in r]

        if fails:
            err = resp[fails[0]]["error"]["message"]
            resp = [r for i, r in enumerate(resp) if i not in fails]
            if not resp and self.disable_retry:
                raise ServiceError(err)

            if "objectIds" in payloads[fails[0]]:
                oids = list(tlz.concat(payloads[i]["objectIds"].split(",") for i in fails))
                self.n_missing = len(oids)

                self.failed_path = Path("cache", f"failed_ids_{self.request_id}.txt")
                with self.failed_path.open("w") as f:
                    f.write("\n".join(oids))

                if not self.disable_retry:
                    self.return_m = bool(payloads[0]["ReturnM"])
                    self.return_geom = bool(payloads[0]["returnGeometry"])
                    self.total_n_features = self.n_features

                    if len(fails) > 1:
                        msg = f"Found {len(fails)} failed requests. Retrying ..."
                        warnings.warn(msg, UserWarning, stacklevel=2)
                    elif len(fails) == 1:
                        msg = "Found 1 failed request. Retrying ..."
                        warnings.warn(msg, UserWarning, stacklevel=2)

                    resp.extend(self.retry_failed_requests())

                    if self.n_missing > 0:
                        msg = " ".join(
                            [
                                f"Total of {self.n_missing} out of {self.total_n_features}",
                                "requested features are not available in the dataset.",
                                "Returning the successfully retrieved features.",
                                "The failed object IDs have been saved in the",
                                f"file {self.failed_path}. The service returned the",
                                "following error message for the failed requests:\n",
                                err,
                            ]
                        )
                        warnings.warn(msg, UserWarning, stacklevel=2)

        return resp

    def get_features(
        self,
        featureids: Iterator[tuple[str, ...]],
        return_m: bool = False,
        return_geom: bool = True,
    ) -> list[dict[str, Any]]:
        """Get features based on the feature IDs.

        Parameters
        ----------
        featureids : list
            List of feature IDs.
        return_m : bool, optional
            Whether to activate the Return M (measure) in the request,
            defaults to ``False``.
        return_geom : bool, optional
            Whether to return the geometry of the feature, defaults to ``True``.

        Returns
        -------
        dict
            (Geo)json response from the web service.
        """
        payloads = [
            {
                "objectIds": ",".join(ids),
                "returnGeometry": str(return_geom).lower(),
                "outSR": str(self.out_sr),
                "outfields": ",".join(self.outfields),
                "ReturnM": str(return_m).lower(),
                "f": self.outformat,
            }
            for ids in featureids
        ]
        if self.request_id is None:
            self.request_id = uuid.uuid4().hex

        resp = self.get_response(self.query_url, payloads, "POST")

        resp = self._cleanup_resp(resp, payloads)
        if not resp:
            raise ZeroMatchedError
        return resp

    def _retry(
        self, return_m: bool, return_geo: bool, partition_fac: float
    ) -> list[dict[str, Any]]:
        """Retry failed requests."""
        with self.failed_path.open() as f:
            oids = [int(i) for i in f.read().splitlines()]

        max_nrecords = self.max_nrecords
        self.max_nrecords = int(max(partition_fac * max_nrecords, 1))
        features = self.get_features(self.partition_oids(oids), return_m, return_geo)
        self.max_nrecords = max_nrecords

        return features

    def retry_failed_requests(self) -> list[dict[str, Any]]:
        """Retry failed requests."""
        retry = self.disable_retry
        self.disable_retry = True
        caching = os.getenv("HYRIVER_CACHE_DISABLE", "false").lower() == "true"
        os.environ["HYRIVER_CACHE_DISABLE"] = "true"

        fac_min = 1.0 / self.max_nrecords
        n_retry = 4
        partition_fac = [(0.5 - fac_min) / (n_retry - 1) * i + fac_min for i in range(n_retry)]
        features = []
        for f in partition_fac[::-1]:
            with contextlib.suppress(ServiceError):
                features.append(self._retry(self.return_m, self.return_geom, f))

        features = list(tlz.concat(features))
        self.n_missing -= len(features)
        self.disable_retry = retry
        os.environ["HYRIVER_CACHE_DISABLE"] = str(caching).lower()
        return features

    def get_response(
        self,
        url: str,
        payloads: list[dict[str, str]],
        method: Literal["get", "GET", "post", "POST"] = "GET",
    ) -> list[dict[str, Any]]:
        """Send payload and get the response."""
        req_key = "params" if method == "GET" else "data"
        try:
            resp = ar.retrieve_json(
                [url] * len(payloads),
                [{req_key: p} for p in payloads],
                request_method=method,
                max_workers=self.max_workers,
            )
            resp = cast("list[dict[str, Any]]", resp)
        except ValueError as ex:
            raise ZeroMatchedError from ex
        return resp

    def __repr__(self) -> str:
        """Print the service configuration."""
        extent = ", ".join(f"{c:.3f}" for c in self.extent) if self.extent else None
        ftypes = ", ".join(self.feature_types.values()) if self.feature_types else None
        return "\n".join(
            [
                "Service configurations:",
                f"    URL: {self.base_url}",
                f"    Layer: {self.valid_layers[str(self.layer)]} ({self.layer})",
                f"    Max record count: {self.max_nrecords}",
                f"    Query format: {self.outformat}",
                f"    Units: {self.units}",
                f"    Extent: ({extent})",
                f"    Feature types: {ftypes}",
            ]
        )


class WMSBase:
    """Base class for accessing a WMS service.

    Parameters
    ----------
    url : str
        The base url for the WMS service e.g., https://www.mrlc.gov/geoserver/mrlc_download/wms
    layers : str or list, optional
        A layer or a list of layers from the service to be downloaded. You can pass an empty
        string to get a list of available layers.
    outformat : str, optional
        The data format to request for data from the service. You can pass an empty
        string to get a list of available output formats.
    version : str, optional
        The WMS service version which should be either 1.1.1 or 1.3.0, defaults to 1.3.0.
    crs : str, int, or pyproj.CRS, optional
        The spatial reference system to be used for requesting the data, defaults to
        ``epsg:4326``.
    validation : bool, optional
        Validate the input arguments from the WMS service, defaults to True. Set this
        to False if you are sure all the WMS settings such as layer and crs are correct
        to avoid sending extra requests.
    """

    def __init__(
        self,
        url: str,
        layers: str | int | Sequence[str | int] = "",
        outformat: str = "",
        version: str = "1.3.0",
        crs: CRSTYPE = 4326,
        validation: bool = True,
    ) -> None:
        self.url = url
        self.outformat = outformat
        self.version = version
        self.crs = crs
        self.outformat = self.outformat.lower()
        if isinstance(layers, (str, int)):
            self.layers = [str(layers)]
        else:
            self.layers = [str(lyr) for lyr in layers]
        self.crs_str = utils.validate_crs(self.crs).lower()
        if self.version not in ("1.1.1", "1.3.0"):
            raise InputValueError("version", ("1.1.1", "1.3.0"))

        if validation:
            self.get_service_options()
            self.validate_wms()

    def get_service_options(self) -> None:
        """Validate input arguments with the WMS service."""
        try:
            wms = WebMapService(self.url, version=self.version)
        except AttributeError as ex:
            raise ServiceUnavailableError(self.url) from ex

        self.available_layer = {wms[lyr].name: wms[lyr].title for lyr in list(wms.contents)}
        self.available_outformat = wms.getOperationByName("GetMap").formatOptions
        self.available_outformat = [f.lower() for f in self.available_outformat]
        self.available_crs = {
            lyr: [s.lower() for s in wms[lyr].crsOptions] for lyr in self.available_layer
        }

    def validate_wms(self) -> None:
        """Validate input arguments with the WMS service."""
        if any(lyr not in self.available_layer for lyr in self.layers):
            raise InputValueError(
                "layers", (f"{n} for {t}" for n, t in self.available_layer.items())
            )

        if self.outformat not in self.available_outformat:
            raise InputValueError("outformat", self.available_outformat)

        if any(self.crs_str not in self.available_crs[lyr] for lyr in self.layers):
            _valid_crss = (
                f"{lyr}: {', '.join(cs)}\n"
                for lyr, cs in self.available_crs.items()
                if lyr in self.layers
            )
            raise InputValueError("CRS", _valid_crss)

    def get_validlayers(self) -> dict[str, str]:
        """Get the layers supported by the WMS service."""
        try:
            wms = WebMapService(self.url, version=self.version)
        except AttributeError as ex:
            raise ServiceUnavailableError(self.url) from ex

        return {wms[lyr].name: wms[lyr].title for lyr in list(wms.contents)}

    def __repr__(self) -> str:
        """Print the services properties."""
        return "\n".join(
            (
                "Connected to the WMS service with the following properties:",
                f"URL: {self.url}",
                f"Version: {self.version}",
                f"Layers: {', '.join(self.layers)}",
                f"Output Format: {self.outformat}",
                f"Output CRS: {self.crs_str}",
            )
        )


class WFSBase:
    """Base class for WFS service.

    Parameters
    ----------
    url : str
        The base url for the WFS service, for examples:
        https://hazards.fema.gov/nfhl/services/public/NFHL/MapServer/WFSServer
    layer : str
        The layer from the service to be downloaded, defaults to None which throws
        an error and includes all the available layers offered by the service.
    outformat : str
        The data format to request for data from the service, defaults to None which
         throws an error and includes all the available format offered by the service.
    version : str, optional
        The WFS service version which should be either ``1.0.0``, ``1.1.0``, or
        ``2.0.0``. Defaults to ``2.0.0``.
    crs : str, int, or pyproj.CRS, optional
        The spatial reference system to be used for requesting the data, defaults to
        ``epsg:4326``.
    read_method : str, optional
        Method for reading the retrieved data, defaults to ``json``. Valid options are
        ``json``, ``binary``, and ``text``.
    max_nrecords : int, optional
        The maximum number of records in a single request to be retrieved from the service,
        defaults to 1000. If the number of requested records is greater than this value,
        the query will be split into multiple requests.
    validation : bool, optional
        Validate the input arguments from the WFS service, defaults to True. Set this
        to False if you are sure all the WFS settings such as layer and crs are correct
        to avoid sending extra requests.
    """

    def __init__(
        self,
        url: str,
        layer: str | None = None,
        outformat: str | None = None,
        version: str = "2.0.0",
        crs: CRSTYPE = 4326,
        read_method: str = "json",
        max_nrecords: int = 1000,
        validation: bool = True,
    ) -> None:
        self.url = url
        self.layer = layer
        self.outformat = outformat
        self.version = version
        self.crs = crs
        self.read_method = read_method
        self.max_nrecords = max_nrecords
        self.crs_str = utils.validate_crs(self.crs)

        if self.version == "2.0.0":
            self.nfeat_key = "numberMatched="
            self.count_key = "count"
        elif self.version == "1.1.0":
            self.nfeat_key = "numberOfFeatures="
            self.count_key = "maxFeatures"
        else:
            raise InputValueError("version", ("1.1.0", "2.0.0"))

        if self.read_method == "text":
            self.retrieve = ar.retrieve_text
        elif self.read_method == "bytes":
            self.retrieve = ar.retrieve_binary
        elif self.read_method == "json":
            self.retrieve = ar.retrieve_json
        else:
            raise InputValueError("read_method", ("json", "binary", "text"))

        self.get_service_options()
        if validation:
            self.validate_wfs()

    def get_service_options(self) -> None:
        """Validate input arguments with the WFS service."""
        try:
            wfs = WebFeatureService(self.url, version=self.version)
        except AttributeError as ex:
            raise ServiceUnavailableError(self.url) from ex

        if wfs is None:
            raise ServiceUnavailableError(self.url)

        self.available_layer = list(wfs.contents)

        wfs_features = wfs.getOperationByName("GetFeature")
        self.available_outformat = wfs_features.parameters["outputFormat"]["values"]
        self.available_outformat = [f.lower() for f in self.available_outformat]

        self.available_crs = {
            lyr: [f"{s.authority.lower()}:{s.code}" for s in wfs[lyr].crsOptions]
            for lyr in self.available_layer
        }

        self.schema = {}
        if self.layer:
            if self.layer not in self.available_layer:
                raise InputValueError("layers", self.available_layer)
            layers = [self.layer]
        else:
            layers = self.available_layer

        for lyr in layers:
            try:
                self.schema[lyr] = wfs.get_schema(lyr)
            except KeyError:
                self.schema[lyr] = None

    def validate_wfs(self) -> None:
        """Validate input arguments with the WFS service."""
        if self.layer is None:
            raise MissingInputError(
                "\n".join(
                    (
                        "The layer argument is missing. The following layers are available:",
                        ", ".join(self.available_layer),
                    )
                )
            )
        if self.layer not in self.available_layer:
            raise InputValueError("layers", self.available_layer)

        if self.outformat not in self.available_outformat:
            raise InputValueError("outformat", self.available_outformat)

        if self.crs_str.lower() not in self.available_crs[self.layer]:
            raise InputValueError("crs", self.available_crs[self.layer])

    def sort_params(
        self, sort_attr: str | None, nfeatures: int, start_index: int
    ) -> dict[str, str]:
        """Get sort parameters for a WFS request."""
        schema = self.schema[self.layer]
        valid_attrs = schema["properties"] if schema else []
        if sort_attr is None:
            if schema is None:
                msg = "sort_attr is None and no id column found in the schema."
                msg += " Please set the sort_attr manually."
                raise MissingInputError(msg)

            sort_attr = next((a for a in valid_attrs if "id" in a.lower()), None)
            if sort_attr is None:
                msg = "sort_attr is None and no id column found in the schema."
                msg += " Please set the sort_attr manually. Available columns are:\n"
                msg += ", ".join(list(valid_attrs))
                raise MissingInputError(msg)

        if valid_attrs and sort_attr not in valid_attrs:
            raise InputValueError("sort_attr", list(valid_attrs))

        if nfeatures <= self.max_nrecords:
            return {
                "startIndex": str(start_index),
                self.count_key: str(nfeatures),
                "sortBy": sort_attr,
            }
        return {
            "startIndex": str(start_index),
            self.count_key: str(self.max_nrecords),
            "sortBy": sort_attr,
        }

    def __repr__(self) -> str:
        """Print the services properties."""
        return "\n".join(
            [
                "Connected to the WFS service with the following properties:",
                f"URL: {self.url}",
                f"Version: {self.version}",
                f"Layer: {self.layer}",
                f"Output Format: {self.outformat}",
                f"Output CRS: {self.crs_str}",
                f"Read Method: {self.read_method}",
            ]
        )



"""Functions for creating unique keys based on web request parameters.

This module is based on the ``aiohttp-client-cache`` package, which is
licensed under the MIT license. See the ``LICENSE`` file for more details.
"""
from __future__ import annotations

import hashlib
from typing import Any, Mapping, Sequence, Union

from multidict import MultiDict
from url_normalize import url_normalize
from yarl import URL

RequestParams = Union[Mapping[Any, Any], Sequence[Any], str, None]
StrOrURL = Union[str, URL]

__all__ = ["create_request_key"]


def _normalize_url_params(url: StrOrURL, params: RequestParams = None) -> URL:
    """Normalize any combination of request parameter formats that aiohttp accepts."""
    if isinstance(url, str):
        url = URL(url)

    # Handle `params` argument, and combine with URL query string if it exists
    if params:
        norm_params = MultiDict(url.query)
        norm_params.extend(url.with_query(params).query)
        url = url.with_query(norm_params)

    # Apply additional normalization and convert back to URL object
    return URL(str(url_normalize(str(url))))


def _encode_dict(data: Any) -> bytes:
    if not data:
        return b""
    if isinstance(data, bytes):
        return data
    elif not isinstance(data, Mapping):
        return str(data).encode()
    item_pairs = [f"{k}={v}" for k, v in sorted((data or {}).items())]
    return "&".join(item_pairs).encode()


def create_request_key(
    method: str,
    url: StrOrURL,
    params: RequestParams = None,
    data: Mapping[str, Any] | None = None,
    json: Mapping[str, Any] | None = None,
) -> str:
    """Create a unique cache key based on request details."""
    # Normalize and filter all relevant pieces of request data
    norm_url = _normalize_url_params(url, params)

    # Create a hash based on the normalized and filtered request
    return hashlib.sha256(
        method.upper().encode() + str(norm_url).encode() + _encode_dict(data) + _encode_dict(json)
    ).hexdigest()



"""Tests for PyGeoOGC package."""
import io
import sys
from dataclasses import fields

import pandas as pd
import pytest
from shapely import LineString, Polygon

import pygeoogc as ogc
from pygeoogc import WFS, WMS, ArcGISRESTful, ServiceURL, utils

DEF_CRS = 4326
ALT_CRS = 4269
GEO_NAT = Polygon(
    [[-69.77, 45.07], [-69.31, 45.07], [-69.31, 45.45], [-69.77, 45.45], [-69.77, 45.07]]
)
GEO_URB = Polygon(
    [
        [-118.72, 34.118],
        [-118.31, 34.118],
        [-118.31, 34.518],
        [-118.72, 34.518],
        [-118.72, 34.118],
    ]
)
COORDS = [(-118.72, 34.118), (-118.31, 34.518)]


class TestREST:
    wbd_url: str = ServiceURL().restful.wbd
    fab_url: str = f"{ServiceURL().restful.nhd_fabric}/1"
    epa_url: str = ServiceURL().restful.nhdplus_epa
    nhd_url: str = ServiceURL().restful.nhdplushr

    def test_byid(self):
        """RESTFul by ID."""
        wbd2 = ArcGISRESTful(self.wbd_url, 1, outfields=["huc2", "name", "areaacres"])
        assert self.wbd_url in wbd2.__repr__()
        huc2 = wbd2.get_features(wbd2.partition_oids(list(range(1, 6))))

        assert len(huc2[0]["features"]) == 5

    def test_geom_point_line(self):
        wbd2 = ArcGISRESTful(self.wbd_url, 1)
        oids = wbd2.oids_bygeom((-70.02580, 44.43280), spatial_relation="esriSpatialRelWithin")
        huc2 = wbd2.get_features(oids)
        point = len(huc2[0]["features"])

        oids = wbd2.oids_bygeom(
            LineString([(-70.02580, 44.43280), (-71.02580, 44.43280)]),
            spatial_relation="esriSpatialRelWithin",
        )
        huc2 = wbd2.get_features(oids)
        line = len(huc2[0]["features"])
        assert point == line

    def test_bygeom(self):
        """RESTFul by geometry."""
        geofab = ArcGISRESTful(self.fab_url, max_workers=4)
        _ = geofab.oids_bygeom(GEO_NAT.bounds)
        oids = geofab.oids_bygeom(GEO_NAT)
        wb_all = geofab.get_features(oids)
        oids = geofab.oids_bygeom(GEO_NAT, sql_clause="areasqkm > 20")
        wb_large = geofab.get_features(oids)
        assert len(wb_all[0]["features"]) - len(wb_large[0]["features"]) == 915

    def test_bymultipoint(self):
        """RESTFul by geometry."""
        geom = [
            (-97.06138, 32.837),
            (-97.06133, 32.836),
            (-97.06124, 32.834),
            (-97.06127, 32.832),
        ]

        service = ArcGISRESTful(self.epa_url, 2, outformat="json")
        oids = service.oids_bygeom(
            geom, geo_crs=ALT_CRS, sql_clause="FTYPE NOT IN (420,428,566)", distance=1500
        )
        resp = service.get_features(oids, return_m=True)
        assert len(resp[0]["features"]) == 3

    @pytest.mark.skip(reason="Service is not available anymore.")
    def test_retry(self):
        rest_url = "/".join(
            [
                "http://cohgims.houstontx.gov/arcgis/rest/services",
                "PUD_Utility/StormwaterUtilities/MapServer/3",
            ]
        )
        rest = ArcGISRESTful(rest_url, verbose=True)
        assert rest_url in rest.__repr__()
        oids = [
            1006322,
            1006323,
            1006324,
            1245206,
            323351,
            324961,
            474077,
            503848,
            584849,
            585138,
        ]
        resp = rest.get_features(rest.partition_oids(oids))

        with rest.client.failed_path.open() as f:
            f_oids = [int(i) for i in f.read().splitlines()]
        assert len(resp) == 3
        assert len(f_oids) == (len(oids) - len(resp))

    def test_bysql(self):
        """RESTFul by SQL filter."""
        hr = ArcGISRESTful(self.nhd_url, 2, outformat="json")
        oids = hr.oids_bysql("nhdplusid IN (50000100239005, 50000100235192)")
        resp = hr.get_features(oids, return_m=True)

        assert len(resp[0]["features"]) == 2

    def test_byfield(self):
        """RESTFul by SQL filter."""
        hr = ArcGISRESTful(self.nhd_url, 2, outformat="json")
        oids = hr.oids_byfield("nhdplusid", ["50000100239005", "50000100235192"])
        resp = hr.get_features(oids)

        assert len(resp[0]["features"]) == 2


def test_retrysession_head():
    url = "https://httpbin.org/image/jpeg"
    with utils.RetrySession() as session:
        resp = session.head(url)
        assert resp.headers["Content-length"] == "35588"


def test_retrysession_get_length():
    url = "https://httpbin.org/image/jpeg"
    with utils.RetrySession(disable=True) as session:
        resp = session.get(url, stream=True)
        assert resp.headers["Content-length"] == "35588"


def test_retrysession_params():
    url = "https://postman-echo.com"
    params = {"params_get": "foo"}
    data = {"data_post": "foo"}
    with utils.RetrySession(disable=True) as session:
        get1 = session.get(f"{url}/get", params=params)
        get2 = session.get(f"{url}/get", payload=params)
        assert get1.json()["args"] == get2.json()["args"]

        post1 = session.post(f"{url}/post", data=data)
        post2 = session.post(f"{url}/post", payload=data)
        assert post1.json()["args"] == post2.json()["args"]


def test_retry_ssl():
    with utils.RetrySession(ssl=False) as session:
        base_url = "/".join(
            (
                "https://gaftp.epa.gov/epadatacommons/ORD",
                "NHDPlusLandscapeAttributes/StreamCat/HydroRegions",
            )
        )
        r = session.get(base_url)
        assert r.status_code == 200


def test_stream():
    url = "http://speedtest.ftp.otenet.gr/files/test100k.db"
    fname = ogc.streaming_download(url)
    assert fname.stat().st_size == 102400

    urls = (
        url,
        "http://speedtest.ftp.otenet.gr/files/test1Mb.db",
    )
    fname = ogc.streaming_download(urls)
    assert fname[0].stat().st_size == 102400
    assert fname[1].stat().st_size == 1048576


@pytest.mark.filterwarnings("ignore:.*Content metadata*.")
class TestWMS:
    wms_url: str = ServiceURL().wms.gebco
    layer: str = "GEBCO_LATEST"

    def test_v111(self):
        """WMS version 1.1.1."""
        wms = WMS(
            self.wms_url,
            layers=self.layer,
            outformat="image/tiff",
            crs=DEF_CRS,
            version="1.1.1",
            validation=False,
        )
        r_dict = wms.getmap_bybox(GEO_NAT.bounds, 20, DEF_CRS)
        assert wms.get_validlayers()[self.layer] == self.layer
        assert sys.getsizeof(r_dict[f"{self.layer}_dd_0_0"]) == 11496857

    def test_bybox(self):
        """WMS by bounding box."""
        wms = WMS(self.wms_url, layers=self.layer, outformat="image/tiff", crs=DEF_CRS)
        assert self.wms_url in wms.__repr__()
        r_dict = wms.getmap_bybox(GEO_NAT.bounds, 20, DEF_CRS, max_px=int(3e6))
        assert sum(sys.getsizeof(r) for r in r_dict.values()) == 11507302
        flist = wms.getmap_bybox(GEO_NAT.bounds, 20, DEF_CRS, max_px=int(3e6), tiff_dir="cache")
        assert sum(f.stat().st_size for f in flist) == 11507170

    def test_valid_crs(self):
        """Get WMS valid CRSs."""
        crs = utils.valid_wms_crs(self.wms_url)
        assert sorted(crs) == ["epsg:3395", "epsg:3857", "epsg:4326"]


class TestWFS:
    wfs: WFS = WFS(
        ServiceURL().wfs.waterdata,
        layer="wmadata:gagesii",
        outformat="csv",
        read_method="text",
        version="2.0.0",
        crs=ALT_CRS,
        max_nrecords=5,
    )

    def to_df(self, resp):
        return pd.concat(pd.read_csv(io.StringIO(r)) for r in resp)

    def test_byid(self):
        """WFS by ID."""
        assert ServiceURL().wfs.waterdata in self.wfs.__repr__()
        stations = [
            "01011000",
            "01013500",
            "01015800",
            "01016500",
            "01017000",
            "01017060",
        ]
        resp = self.wfs.getfeature_byid("staid", stations)
        df = self.to_df(resp)
        assert set(df.staid.unique()) == {int(s) for s in stations}

    def test_bygeom(self):
        """WFS by geometry."""
        resp = self.wfs.getfeature_bygeom(GEO_URB, geo_crs=DEF_CRS, always_xy=False)
        assert self.to_df(resp).shape[0] == 7

    def test_bybox(self):
        """WFS by bounding box."""
        bbox = GEO_URB.bounds
        resp = self.wfs.getfeature_bybox(bbox, box_crs=DEF_CRS, always_xy=True)
        assert self.to_df(resp).shape[0] == 7

    def test_byfilter(self):
        """WFS by CQL filter."""
        resp = self.wfs.getfeature_byfilter("staid LIKE '010315%'")
        assert self.to_df(resp).shape[0] == 2

    def test_wfs110(self):
        """WFS 1.1.0 by geom."""
        resp = WFS(
            ServiceURL().wfs.waterdata,
            layer="wmadata:gagesii",
            outformat="json",
            version="1.1.0",
            crs=ALT_CRS,
        ).getfeature_bygeom(GEO_URB, geo_crs=DEF_CRS, always_xy=False)
        assert len(resp[0]["features"]) == 7


def test_decompose():
    """Bounding box decomposition."""
    bboxs = utils.bbox_decompose(GEO_URB.bounds, 10)
    assert bboxs[0][-1] == 2828


@pytest.mark.parametrize(
    ("geo", "gtype", "expected"),
    [
        (COORDS, "coords", -287.707),
        (GEO_URB, "geometry", -362.099),
        (GEO_URB.bounds, "bounds", -365.403),
    ],
)
def test_matchcrs(geo, gtype, expected):
    """Match CRS."""
    matched = utils.match_crs(geo, DEF_CRS, 2149)
    if gtype == "coords":
        val = matched[-1][0]
    elif gtype == "geometry":
        val = matched.centroid.x
    elif gtype == "bounds":
        val = matched[0]
    assert (val * 1e-4 - expected) < 1e-3


def test_esriquery():
    """ESRI geometry query builder."""
    point = utils.ESRIGeomQuery(COORDS[0], wkid=DEF_CRS).point()
    line = utils.ESRIGeomQuery(LineString(COORDS), wkid=DEF_CRS).polyline()
    assert point["geometryType"] == "esriGeometryPoint"
    assert line["geometryType"] == "esriGeometryPolyline"


def test_urls():
    """Service URLs."""
    total_urls = 0
    service_url_class = ServiceURL()
    for field in fields(service_url_class):
        nested_class = getattr(service_url_class, field.name)
        total_urls += sum(
            isinstance(getattr(nested_class, f.name), str) for f in fields(nested_class)
        )
    assert total_urls == 38


def test_show_versions():
    """Show versions."""
    f = io.StringIO()
    ogc.show_versions(file=f)
    assert "SYS INFO" in f.getvalue()



"""Some utilities for PyGeoOGC."""
# pyright: reportGeneralTypeIssues=false
from __future__ import annotations

import contextlib
import itertools
import math
import os
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generator,
    Literal,
    Mapping,
    Sequence,
    TypeVar,
    Union,
    cast,
    overload,
)

import cytoolz.curried as tlz
import defusedxml.ElementTree as ETree
import joblib
import pyproj
import requests
import ujson
import urllib3
from pyproj.exceptions import CRSError as ProjCRSError
from requests.adapters import HTTPAdapter
from requests.exceptions import RequestException
from requests_cache import CachedSession, Response
from requests_cache.backends.sqlite import SQLiteCache
from shapely import LineString, MultiLineString, MultiPoint, MultiPolygon, Point, Polygon, ops
from shapely import box as shapely_box
from urllib3.exceptions import InsecureRequestWarning

import async_retriever as ar
from pygeoogc import cache_keys
from pygeoogc.exceptions import InputTypeError, InputValueError, ServiceError

if TYPE_CHECKING:
    from typing_extensions import Self

    CRSTYPE = Union[int, str, pyproj.CRS]
    GEOM = TypeVar(
        "GEOM",
        Point,
        MultiPoint,
        Polygon,
        MultiPolygon,
        LineString,
        MultiLineString,
        "tuple[float, float, float, float]",
        "list[tuple[float, float]]",
    )
BOX_ORD = "(west, south, east, north)"
MAX_CONN = 10
CHUNK_SIZE = int(100 * 1024 * 1024)  # 100 MB
EXPIRE_AFTER = 60 * 60 * 24 * 7  # 1 week
__all__ = ["RetrySession", "traverse_json", "streaming_download", "match_crs", "validate_crs"]

warnings.filterwarnings("ignore", message=".*too short worker timeout.*")


def check_response(resp: str) -> str:
    """Extract error message from a response, if any."""
    try:
        root = ETree.fromstring(resp)
    except ETree.ParseError:
        return resp
    else:
        try:
            return str(root[-1][0].text).strip()
        except IndexError:
            try:
                return str(root[-1].text).strip()
            except IndexError:
                return str(root.text).strip()


class RetrySession:
    """Configures the passed-in session to retry on failed requests.

    Notes
    -----
    The fails can be due to connection errors, specific HTTP response
    codes and 30X redirections. The code was originally based on:
    https://github.com/bustawin/retry-requests

    Parameters
    ----------
    retries : int, optional
        The number of maximum retries before raising an exception, defaults to 5.
    backoff_factor : float, optional
        A factor used to compute the waiting time between retries, defaults to 0.5.
    status_to_retry : tuple, optional
        A tuple of status codes that trigger the reply behaviour, defaults to (500, 502, 504).
    prefixes : tuple, optional
        The prefixes to consider, defaults to ("http://", "https://")
    cache_name : str, optional
        Path to a folder for caching the session, default to None which uses
        system's temp directory.
    expire_after : int, optional
        Expiration time for the cache in seconds, defaults to -1 (never expire).
    disable : bool, optional
        If ``True`` temporarily disable caching request/responses, defaults to ``False``.
    ssl : bool, optional
        If ``True`` verify SSL certificates, defaults to ``True``.
    """

    def __init__(
        self,
        retries: int = 3,
        backoff_factor: float = 0.3,
        status_to_retry: tuple[int, ...] = (500, 502, 504),
        prefixes: tuple[str, ...] = ("https://",),
        cache_name: str | Path | None = None,
        expire_after: int = EXPIRE_AFTER,
        disable: bool = False,
        ssl: bool = True,
    ) -> None:
        if cache_name is None:
            self.cache_name = Path(
                os.getenv("HYRIVER_CACHE_NAME_HTTP", Path("cache", "http_cache.sqlite"))
            )
        else:
            self.cache_name = Path(cache_name)
        self.cache_name.parent.mkdir(exist_ok=True, parents=True)

        self.expire_after = expire_after
        if self.expire_after == EXPIRE_AFTER:
            self.expire_after = int(os.getenv("HYRIVER_CACHE_EXPIRE", EXPIRE_AFTER))

        self.ssl_cert = os.getenv("HYRIVER_SSL_CERT")

        self.disable = disable

        if not ssl:
            urllib3.disable_warnings(InsecureRequestWarning)
            self.session.verify = False

        adapter = HTTPAdapter(
            max_retries=urllib3.Retry(
                total=retries,
                read=retries,
                connect=retries,
                backoff_factor=backoff_factor,
                status_forcelist=status_to_retry,
                allowed_methods=None,
            )
        )
        for prefix in prefixes:
            self.session.mount(prefix, adapter)

    @property
    def disable(self) -> bool:
        """Disable caching request/responses."""
        return self._disable

    @disable.setter
    def disable(self, value: bool) -> None:
        self._disable = value
        if not self._disable:
            self._disable = os.getenv("HYRIVER_CACHE_DISABLE", "false").lower() == "true"

        if self._disable:
            self.session = requests.Session()
        else:
            backend = SQLiteCache(self.cache_name, fast_save=True, timeout=1)
            self.session = CachedSession(expire_after=self.expire_after, backend=backend)
        self.session.cert = self.ssl_cert

    def get(
        self,
        url: str,
        payload: Mapping[str, Any] | None = None,
        params: Mapping[str, Any] | None = None,
        headers: Mapping[str, Any] | None = None,
        stream: bool | None = None,
    ) -> Response:
        """Retrieve data from a url by GET and return the Response."""
        if stream and not self.disable:
            msg = ". ".join(
                (
                    "Streaming is not supported with caching enabled",
                    "Reinstantiate the class like so: RetrySession(disable=True).",
                )
            )
            raise ValueError(msg)
        params = params or payload
        resp = self.session.get(url, params=params, headers=headers, stream=stream)
        try:
            resp.raise_for_status()
        except RequestException as ex:
            raise ServiceError(check_response(resp.text)) from ex
        else:
            return resp

    def post(
        self,
        url: str,
        payload: Mapping[str, Any] | None = None,
        data: Mapping[str, Any] | None = None,
        json: Mapping[str, Any] | None = None,
        headers: Mapping[str, Any] | None = None,
        stream: bool | None = None,
    ) -> Response:
        """Retrieve data from a url by POST and return the Response."""
        if stream and not self.disable:
            msg = ". ".join(
                (
                    "Streaming is not supported with caching enabled",
                    "Reinstantiate the class like so: RetrySession(disable=True).",
                )
            )
            raise ValueError(msg)
        data = data or payload
        resp = self.session.post(url, data=data, json=json, headers=headers, stream=stream)
        try:
            resp.raise_for_status()
        except RequestException as ex:
            raise ServiceError(check_response(resp.text)) from ex
        else:
            return resp

    def head(
        self,
        url: str,
        params: Mapping[str, Any] | None = None,
        data: Mapping[str, Any] | None = None,
        json: Mapping[str, Any] | None = None,
        headers: Mapping[str, Any] | None = None,
    ) -> Response:
        """Retrieve data from a url by POST and return the Response."""
        resp = self.session.head(url, data=data, params=params, json=json, headers=headers)
        try:
            resp.raise_for_status()
        except RequestException as ex:
            raise ServiceError(check_response(resp.text)) from ex
        else:
            return resp

    def close(self) -> None:
        """Close the session."""
        self.session.close()

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: object) -> None:
        self.close()


def _prepare_requests_args(
    urls: list[str] | str,
    kwds: list[dict[str, dict[Any, Any]]] | dict[str, dict[Any, Any]] | None,
    method: str,
    fnames: str | Path | Sequence[str | Path] | None,
    root_dir: str | Path | None,
    file_prefix: str,
    file_extention: str,
) -> tuple[
    tuple[str, ...], tuple[dict[str, None | dict[Any, Any]], ...], Generator[Path, None, None]
]:
    """Get url and kwds for streaming download."""
    url_list = (urls,) if isinstance(urls, str) else tuple(urls)

    if kwds is None:
        if method == "GET":
            kwd_list = ({"params": None},) * len(url_list)
        else:
            kwd_list = ({"data": None},) * len(url_list)
    else:
        kwd_list = (kwds,) if isinstance(kwds, dict) else tuple(kwds)
    key_list = set(itertools.chain.from_iterable(k.keys() for k in kwd_list))
    valid_keys = ("params", "data", "json", "headers")
    if any(k not in valid_keys for k in key_list):
        raise InputValueError("kwds", valid_keys)

    if len(url_list) != len(kwd_list):
        raise InputTypeError("urls/kwds", "list of same length")

    fex = file_extention.replace(".", "")

    if fnames is None:
        if root_dir is None:
            root_dir = os.getenv("HYRIVER_CACHE_NAME", str(Path("cache", "tmp")))
            root_dir = Path(root_dir).parent
        else:
            root_dir = Path(root_dir)
        root_dir.mkdir(exist_ok=True, parents=True)
        files = (
            Path(root_dir, f"{file_prefix}{cache_keys.create_request_key(method, u, **p)}.{fex}")
            for u, p in zip(url_list, kwd_list)
        )
    else:
        f_list = (fnames,) if isinstance(fnames, (str, Path)) else tuple(fnames)
        if len(url_list) != len(f_list):
            raise InputTypeError("urls/fnames", "lists of same length")
        files = (Path(f) for f in f_list)
    url_list = cast("tuple[str, ...]", url_list)
    kwd_list = cast("tuple[dict[str, None | dict[Any, Any]], ...]", kwd_list)
    return url_list, kwd_list, files


def _download(
    session_func: Callable[[str], Response],
    url: str,
    kwd: Mapping[str, None | dict[Any, Any]],
    fname: Path,
    chunk_size: int,
) -> Path:
    """Download a single file."""
    resp = session_func(url, **kwd)
    fsize = int(resp.headers.get("Content-Length", -1))
    if not fname.exists() or fname.stat().st_size != fsize:
        fname.parent.mkdir(exist_ok=True, parents=True)
        with fname.open("wb") as f:
            f.writelines(resp.iter_content(chunk_size))
    return fname


@overload
def streaming_download(
    urls: str,
    kwds: dict[str, dict[Any, Any]] | None = None,
    fnames: str | Path | None = None,
    root_dir: str | Path | None = None,
    file_prefix: str = "",
    file_extention: str = "",
    method: str = "GET",
    ssl: bool = True,
    chunk_size: int = CHUNK_SIZE,
    n_jobs: int = MAX_CONN,
) -> Path:
    ...


@overload
def streaming_download(
    urls: list[str],
    kwds: list[dict[str, dict[Any, Any]]] | None = None,
    fnames: Sequence[str | Path] | None = None,
    root_dir: str | Path | None = None,
    file_prefix: str = "",
    file_extention: str = "",
    method: str = "GET",
    ssl: bool = True,
    chunk_size: int = CHUNK_SIZE,
    n_jobs: int = MAX_CONN,
) -> list[Path]:
    ...


def streaming_download(
    urls: list[str] | str,
    kwds: list[dict[str, dict[Any, Any]]] | dict[str, dict[Any, Any]] | None = None,
    fnames: str | Path | Sequence[str | Path] | None = None,
    root_dir: str | Path | None = None,
    file_prefix: str = "",
    file_extention: str = "",
    method: str = "GET",
    ssl: bool = True,
    chunk_size: int = CHUNK_SIZE,
    n_jobs: int = MAX_CONN,
) -> Path | list[Path]:
    """Download and store files in parallel from a list of URLs/Keywords.

    Notes
    -----
    This function runs asynchronously in parallel using ``n_jobs`` threads.

    Parameters
    ----------
    urls : tuple or list
        A list of URLs to download.
    kwds : tuple or list, optional
        A list of keywords associated with each URL, e.g.,
        ({"params": ..., "headers": ...}, ...). Defaults to ``None``.
    fnames : tuple or list, optional
        A list of filenames associated with each URL, e.g.,
        ("file1.zip", ...). Defaults to ``None``. If not provided,
        random unique filenames will be generated based on
        URL and keyword pairs.
    root_dir : str or Path, optional
        Root directory to store the files, defaults to ``None`` which
        uses HyRiver's cache directory. Note that you should either
        provide ``root_dir`` or ``fnames``. If both are provided,
        ``root_dir`` will be ignored.
    file_prefix : str, optional
        Prefix to add to filenames when storing the files, defaults
        to ``None``, i.e., no prefix. This argument will be only be
        used if ``fnames`` is not passed.
    file_extention : str, optional
        Extension to use for storing the files, defaults to ``None``,
        i.e., no extension if ``fnames`` is not provided otherwise. This
        argument will be only be used if ``fnames`` is not passed.
    method : str, optional
        HTTP method to use, i.e, ``GET`` or ``POST``, by default "GET".
    ssl : bool, optional
        Whether to use SSL verification, defaults to ``True``.
    chunk_size : int, optional
        Chunk size to use when downloading, defaults to 100 * 1024 * 1024
        i.e., 100 MB.
    n_jobs: int, optional
        The maximum number of concurrent downloads, defaults to 10.

    Returns
    -------
    list
        A list of ``pathlib.Path`` objects associated with URLs in the
        same order.
    """
    method = method.upper()
    valid_methods = ("GET", "POST")
    if method not in valid_methods:
        raise InputValueError("method", valid_methods)

    url_list, kwd_list, files = _prepare_requests_args(
        urls, kwds, method, fnames, root_dir, file_prefix, file_extention
    )

    session = RetrySession(disable=True, ssl=ssl)
    if method == "GET":
        func = tlz.partial(session.get, stream=True)
    else:
        func = tlz.partial(session.post, stream=True)

    n_jobs = min(n_jobs, len(url_list))
    fpaths = joblib.Parallel(n_jobs=n_jobs, prefer="threads")(
        joblib.delayed(_download)(func, u, k, f, chunk_size)
        for u, k, f in zip(url_list, kwd_list, files)
    )
    fpaths = cast("list[Path]", fpaths)
    session.close()
    if isinstance(urls, str):
        return fpaths[0]
    return fpaths


def traverse_json(json_data: dict[str, Any] | list[dict[str, Any]], ipath: list[str]) -> list[Any]:
    """Extract an element from a JSON-like object along a specified ipath.

    This function is based on
    `bcmullins <https://bcmullins.github.io/parsing-json-python/>`__.

    Parameters
    ----------
    json_data : dict or list of dicts
        The input json dictionary.
    ipath : list
        The ipath to the requested element.

    Returns
    -------
    list
        The sub-items founds in the JSON.

    Examples
    --------
    >>> data = [
    ...     {"employees": [
    ...         {"name": "Alice", "role": "dev", "nbr": 1},
    ...         {"name": "Bob", "role": "dev", "nbr": 2},
    ...         ],},
    ...     {"firm": {"name": "Charlie's Waffle Emporium", "location": "CA"}},
    ... ]
    >>> traverse_json(data, ["employees", "name"])
    [['Alice', 'Bob'], [None]]
    """

    def extract(
        sub_items: list[Any] | dict[str, Any] | None,
        path: list[str],
        ind: int,
        arr: list[Any],
    ) -> list[Any]:
        key = path[ind]
        if isinstance(sub_items, dict):
            if key in sub_items:
                if ind + 1 == len(path):
                    arr.append(sub_items[key])
                else:
                    extract(sub_items[key], path, ind + 1, arr)
            else:
                arr.append(None)
        elif isinstance(sub_items, list):
            if not sub_items:
                arr.append(None)
            else:
                for i in sub_items:
                    extract(i, path, ind, arr)
        else:
            arr.append(None)
        return arr

    if isinstance(json_data, dict):
        return extract(json_data, ipath, 0, [])

    outer_arr = [extract(item, ipath, 0, []) for item in json_data]
    return outer_arr


@dataclass
class ESRIGeomQuery:
    """Generate input geometry query for ArcGIS RESTful services.

    Parameters
    ----------
    geometry : tuple or Polygon or Point or LineString
        The input geometry which can be a point (x, y), a list of points [(x, y), ...],
        bbox (xmin, ymin, xmax, ymax), or a Shapely's Polygon.
    wkid : int
        The Well-known ID (WKID) of the geometry's spatial reference e.g., for EPSG:4326,
        4326 should be passed. Check
        `ArcGIS <https://developers.arcgis.com/rest/services-reference/geographic-coordinate-systems.htm>`__
        for reference.
    """

    geometry: (
        tuple[float, float]
        | list[tuple[float, float]]
        | tuple[float, float, float, float]
        | Polygon
        | LineString
    )
    wkid: int

    def _get_payload(
        self,
        geo_type: Literal[
            "esriGeometryPoint",
            "esriGeometryMultipoint",
            "esriGeometryEnvelope",
            "esriGeometryPolygon",
            "esriGeometryPolyline",
        ],
        geo_json: dict[str, Any],
    ) -> Mapping[str, str]:
        """Generate a request payload based on ESRI template.

        Parameters
        ----------
        geo_type : str
            Type of the input geometry
        geo_json : dict
            Geometry in GeoJson format.

        Returns
        -------
        dict
            An ESRI geometry payload.
        """
        esri_json = ujson.dumps({**geo_json, "spatialRelference": {"wkid": str(self.wkid)}})
        return {
            "geometryType": geo_type,
            "geometry": esri_json,
            "inSR": str(self.wkid),
        }

    def point(self) -> Mapping[str, str]:
        """Query for a point."""
        try:
            point = Point(*self.geometry)
        except (TypeError, AttributeError, ValueError) as ex:
            raise InputTypeError("geometry", "tuple", "(x, y)") from ex
        geo_type = "esriGeometryPoint"
        geo_json = {"x": point.x, "y": point.y}
        return self._get_payload(geo_type, geo_json)

    def multipoint(self) -> Mapping[str, str]:
        """Query for a multi-point."""
        try:
            mp = MultiPoint(self.geometry)
        except (TypeError, AttributeError, ValueError) as ex:
            raise InputTypeError("geometry", "list of tuples", "[(x, y), ...]") from ex
        geo_type = "esriGeometryMultipoint"
        geo_json = {"points": [[p.x, p.y] for p in mp.geoms]}
        return self._get_payload(geo_type, geo_json)

    def bbox(self) -> Mapping[str, str]:
        """Query for a bbox."""
        try:
            bbox = shapely_box(*self.geometry)
        except (TypeError, AttributeError, ValueError) as ex:
            raise InputTypeError("geometry", "tuple", BOX_ORD) from ex
        geo_type = "esriGeometryEnvelope"
        geo_json = dict(zip(("xmin", "ymin", "xmax", "ymax"), bbox.bounds))
        return self._get_payload(geo_type, geo_json)

    def polygon(self) -> Mapping[str, str]:
        """Query for a polygon."""
        if isinstance(self.geometry, Polygon):
            geo_type = "esriGeometryPolygon"
            geo_json = {"rings": [[[x, y] for x, y in zip(*self.geometry.exterior.coords.xy)]]}
            return self._get_payload(geo_type, geo_json)

        raise InputTypeError("geometry", "Polygon")

    def polyline(self) -> Mapping[str, str]:
        """Query for a polyline."""
        if isinstance(self.geometry, LineString):
            geo_type = "esriGeometryPolyline"
            geo_json = {"paths": [[[x, y] for x, y in zip(*self.geometry.coords.xy)]]}
            return self._get_payload(geo_type, geo_json)

        raise InputTypeError("geometry", "LineString")


def match_crs(geom: GEOM, in_crs: CRSTYPE, out_crs: CRSTYPE) -> GEOM:
    """Reproject a geometry to another CRS.

    Parameters
    ----------
    geom : list or tuple or geometry
        Input geometry which could be a list of coordinates such as ``[(x1, y1), ...]``,
        a bounding box like so ``(xmin, ymin, xmax, ymax)``, or any valid ``shapely``'s
        geometry such as ``Polygon``, ``MultiPolygon``, etc..
    in_crs : str, int, or pyproj.CRS
        Spatial reference of the input geometry
    out_crs : str, int, or pyproj.CRS
        Target spatial reference

    Returns
    -------
    same type as the input geometry
        Transformed geometry in the target CRS.

    Examples
    --------
    >>> from shapely import Point
    >>> point = Point(-7766049.665, 5691929.739)
    >>> match_crs(point, 3857, 4326).xy
    (array('d', [-69.7636111130079]), array('d', [45.44549114818127]))
    >>> bbox = (-7766049.665, 5691929.739, -7763049.665, 5696929.739)
    >>> match_crs(bbox, 3857, 4326)
    (-69.7636111130079, 45.44549114818127, -69.73666165448431, 45.47699468552394)
    >>> coords = [(-7766049.665, 5691929.739)]
    >>> match_crs(coords, 3857, 4326)
    [(-69.7636111130079, 45.44549114818127)]
    """
    project = pyproj.Transformer.from_crs(in_crs, out_crs, always_xy=True).transform

    if isinstance(
        geom,
        (
            Polygon,
            LineString,
            MultiLineString,
            MultiPolygon,
            Point,
            MultiPoint,
        ),
    ):
        if pyproj.CRS(in_crs) == pyproj.CRS(out_crs):
            return geom
        return ops.transform(project, geom)

    with contextlib.suppress(TypeError, AttributeError, ValueError):
        if len(geom) > 4:
            raise TypeError
        if pyproj.CRS(in_crs) == pyproj.CRS(out_crs):
            bbox = shapely_box(*geom)
        else:
            bbox = cast("Polygon", ops.transform(project, shapely_box(*geom)))
        return tuple(float(p) for p in bbox.bounds)

    with contextlib.suppress(TypeError, AttributeError, ValueError):
        if pyproj.CRS(in_crs) == pyproj.CRS(out_crs):
            point = Point(geom)
        else:
            point = cast("Point", ops.transform(project, Point(geom)))
        return [(float(point.x), float(point.y))]

    with contextlib.suppress(TypeError, AttributeError, ValueError):
        if pyproj.CRS(in_crs) == pyproj.CRS(out_crs):
            mp = MultiPoint(geom)
        else:
            mp = cast("MultiPoint", ops.transform(project, MultiPoint(geom)))
        return [(float(p.x), float(p.y)) for p in mp.geoms]

    gtypes = " ".join(
        (
            "a list of coordinates such as [(x1, y1), ...],",
            "a bounding box like so (xmin, ymin, xmax, ymax),",
            "or any valid shapely's geometry.",
        )
    )
    raise InputTypeError("geom", gtypes)


def esri_query(
    geom: GEOM,
    geo_crs: CRSTYPE,
    out_crs: CRSTYPE,
) -> Mapping[str, str]:
    """Generate geometry queries based on ESRI template."""
    geom = match_crs(geom, geo_crs, out_crs)
    wkid = cast("int", pyproj.CRS(out_crs).to_epsg())

    with contextlib.suppress(TypeError, AttributeError, ValueError):
        geom = Point(geom)

    with contextlib.suppress(TypeError, AttributeError, ValueError):
        geom = MultiPoint(geom)

    if isinstance(geom, Point):
        return ESRIGeomQuery((geom.x, geom.y), wkid).point()

    if isinstance(geom, MultiPoint):
        return ESRIGeomQuery([(g.x, g.y) for g in geom.geoms], wkid).multipoint()

    if isinstance(geom, Polygon):
        return ESRIGeomQuery(geom, wkid).polygon()

    if isinstance(geom, LineString):
        return ESRIGeomQuery(geom, wkid).polyline()

    with contextlib.suppress(InputTypeError):
        return ESRIGeomQuery(geom, wkid).bbox()

    raise InputTypeError("geom", "LineString, Polygon, Point, MultiPoint, tuple")


def check_bbox(bbox: tuple[float, float, float, float]) -> None:
    """Check if an input inbox is a tuple of length 4."""
    try:
        _ = shapely_box(*bbox)
    except (TypeError, AttributeError, ValueError) as ex:
        raise InputTypeError("bbox", "tuple", BOX_ORD) from ex


def bbox_decompose(
    bbox: tuple[float, float, float, float],
    resolution: float,
    box_crs: CRSTYPE = 4326,
    max_px: int = 8000000,
) -> list[tuple[tuple[float, float, float, float], str, int, int]]:
    r"""Split the bounding box vertically for WMS requests.

    Parameters
    ----------
    bbox : tuple
        A bounding box; (west, south, east, north)
    resolution : float
        The target resolution for a WMS request in meters.
    box_crs : str, int, or pyproj.CRS, optional
        The spatial reference of the input bbox, default to ``epsg:4326``.
    max_px : int, optional
        The maximum allowable number of pixels (width x height) for a WMS requests,
        defaults to 8 million based on some trial-and-error.

    Returns
    -------
    list of tuples
        Each tuple includes the following elements:

        * Tuple of px_tot 4 that represents a bounding box (west, south, east, north) of a cell,
        * A label that represents cell ID starting from bottom-left to top-right, for example a
          2x2 decomposition has the following labels::

          |---------|---------|
          |         |         |
          |   0_1   |   1_1   |
          |         |         |
          |---------|---------|
          |         |         |
          |   0_0   |   1_0   |
          |         |         |
          |---------|---------|

        * Raster width of a cell,
        * Raster height of a cell.

    """
    check_bbox(bbox)

    geod = pyproj.Geod(ellps="GRS80")

    west, south, east, north = bbox

    xmin, ymin, xmax, ymax = match_crs(bbox, box_crs, 4326)

    x_dist = geod.geometry_length(LineString([(xmin, ymin), (xmax, ymin)]))
    y_dist = geod.geometry_length(LineString([(xmin, ymin), (xmin, ymax)]))
    width = int(math.ceil(x_dist / resolution))
    height = int(math.ceil(y_dist / resolution))

    if width * height <= max_px:
        bboxs = [(bbox, "0_0", width, height)]

    n_px = int(math.sqrt(max_px))

    def _split_directional(low: float, high: float, px_tot: int) -> tuple[list[int], list[float]]:
        npt = [n_px for _ in range(int(px_tot / n_px))] + [px_tot % n_px]
        xd = abs(high - low)
        dx = [xd * n / sum(npt) for n in npt]
        xs = [low + d for d in itertools.accumulate(dx)]
        xs.insert(0, low)
        return npt, xs

    nw, xs = _split_directional(west, east, width)
    nh, ys = _split_directional(south, north, height)

    bboxs = []
    for j, h in enumerate(nh):
        for i, w in enumerate(nw):
            bx_crs = (xs[i], ys[j], xs[i + 1], ys[j + 1])
            bboxs.append((bx_crs, f"{i}_{j}", w, h))
    return bboxs


def validate_crs(crs: CRSTYPE) -> str:
    """Validate a CRS.

    Parameters
    ----------
    crs : str, int, or pyproj.CRS
        Input CRS.

    Returns
    -------
    str
        Validated CRS as a string.
    """
    try:
        return pyproj.CRS(crs).to_string()
    except ProjCRSError as ex:
        raise InputTypeError("crs", "a valid CRS") from ex


def valid_wms_crs(url: str) -> list[str]:
    """Get valid CRSs from a WMS service version 1.3.0."""
    ns = "http://www.opengis.net/wms"

    def get_path(tag_list: list[str]) -> str:
        return f"/{{{ns}}}".join(["", *tag_list])[1:]

    kwds = {"params": {"service": "wms", "request": "GetCapabilities"}}
    root = ETree.fromstring(ar.retrieve_text([url], [kwds], ssl=False)[0])
    return [t.text.lower() for t in root.findall(get_path(["Capability", "Layer", "CRS"]))]
